/***************************************************************************
                           drivedialog.cpp
                           -------------------------------
    copyright            : (C) 2010 Guido Scholz
    e-mail               : guido.scholz@bayernline.de
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 This file provides an user interface to change drive configuration
 ***************************************************************************/

#include <QLabel>
#include <QLayout>
#include <QMessageBox>
#include <QDialogButtonBox>

#include "drivedialog.h"
#include "preferences.h"
#include "resources.h"
#include "srcpmessage.h"

enum {
    MINMMPORT = 0,
    MAXMMPORT = 1,
    MINDCCPORT = 0,
    MAXDCCPORT = 1,
    MINSVPORT = 0,
    MAXSVPORT = 64535, // no limits
    MINSXPORT = 1,
    MAXSXPORT = 8
};



DriveDialog::DriveDialog(QWidget* parent): QDialog(parent)
{
    setObjectName("DriveDialog");
    setWindowTitle(tr("Edit drive"));

    /*Layout to separate OK Cancel buttons from the upper rest*/
    QBoxLayout* baseLayout = new QVBoxLayout(this);
    baseLayout->setSpacing(10);

    /*Layout to separate left and right groupboxes*/
    QBoxLayout* leftRightLayout = new QHBoxLayout();
    baseLayout->addLayout(leftRightLayout);
    leftRightLayout->setSpacing(10);

    /*left groupbox with protocol data*/
    QGroupBox* protocolGB = new QGroupBox(tr("Protocol"));
    protocolGB->setObjectName("protocolGB");
    leftRightLayout->addWidget(protocolGB);

    QVBoxLayout* protocolGBL = new QVBoxLayout();
    protocolGB->setLayout(protocolGBL);

    protocolBG = new QButtonGroup(protocolGBL);
    protocolBG->setExclusive(true);

    QRadioButton* mmRB = new QRadioButton(tr("&Maerklin/Motorola"),
            protocolGB);
    protocolBG->addButton(mmRB, GaMessage::proMM);
    protocolGBL->addWidget(mmRB);

    QRadioButton* dccRB = new QRadioButton(tr("&NMRA/DCC"), protocolGB);
    protocolBG->addButton(dccRB, GaMessage::proDCC);
    protocolGBL->addWidget(dccRB);

    QRadioButton* selRB = new QRadioButton(tr("Selectri&x"), protocolGB);
    protocolBG->addButton(selRB, GaMessage::proSelectrix);
    protocolGBL->addWidget(selRB);

    QRadioButton* serverRB = new QRadioButton(tr("Protocol by Ser&ver"),
            protocolGB);
    protocolBG->addButton(serverRB, GaMessage::proServer);
    protocolGBL->addWidget(serverRB);
    protocolGBL->addStretch();

    connect(protocolBG, SIGNAL(buttonClicked(int)),
            this, SLOT(slotProtocolChanged(int)));

    /*right decoder data group box*/
    QGroupBox* decoderGB = new QGroupBox(tr("Decoder"), this);
    decoderGB->setObjectName("decoderGB");
    leftRightLayout->addWidget(decoderGB);

    QVBoxLayout* decoderGBL = new QVBoxLayout();
    decoderGB->setLayout(decoderGBL);

    /*line with resettime */
    QHBoxLayout* resetLayout = new QHBoxLayout();
    decoderGBL->addLayout(resetLayout);

    QLabel* labelTime = new QLabel(tr("Reset &after (ms):"), decoderGB);
    resetLayout->addWidget(labelTime);
    resetLayout->addStretch();

    activeTimeSB = new QSpinBox(decoderGB);
    activeTimeSB->setWrapping(true);
    activeTimeSB->setMinimum(50);
    activeTimeSB->setMaximum(2000);
    activeTimeSB->setSingleStep(50);
    resetLayout->addWidget(activeTimeSB);
    labelTime->setBuddy(activeTimeSB);

    QGridLayout* decdataLayout = new QGridLayout();
    decdataLayout->setObjectName("decdataLayout");
    decoderGBL->addLayout(decdataLayout);

    /*line with srcp bus 1 */
    QLabel* srcpBus1Label = new QLabel(tr("S&RCP-Bus:"), decoderGB);
    decdataLayout->addWidget(srcpBus1Label, 0, 0);
    srcpBus1LE = new QLineEdit(decoderGB);
    srcpBus1LE->setObjectName("srcpBus1LE");
    srcpBus1LE->setMaxLength(4);
    srcpBus1LE->setMaximumWidth(LEMAXWIDTH);
    decdataLayout->addWidget(srcpBus1LE, 0, 1);
    srcpBus1Label->setBuddy(srcpBus1LE);

    /*line with address 1 */
    QLabel* address1Lbl = new QLabel(tr("A&ddress:"), decoderGB);
    decdataLayout->addWidget(address1Lbl, 1, 0);
    address1LE = new QLineEdit(decoderGB);
    address1LE->setObjectName("address_1");
    decdataLayout->addWidget(address1LE, 1, 1);
    address1LE->setMaxLength(4);       // address length of NA protocol
    address1LE->setMaximumWidth(LEMAXWIDTH);
    addressVdt = new QIntValidator(0, MAX_GADCC, this);
    address1LE->setValidator(addressVdt);
    address1Lbl->setBuddy(address1LE);

    /*line with port 1 spinbox */
    port1Label = new QLabel(tr("&Port:"), decoderGB);
    decdataLayout->addWidget(port1Label, 2, 0);

    port1SB = new QSpinBox(decoderGB);
    port1SB->setObjectName("port1SB");
    port1SB->setWrapping(true);
    port1SB->setMinimum(0);
    port1SB->setMaximum(1);
    decdataLayout->addWidget(port1SB, 2, 1);
    port1Label->setBuddy(port1SB);

    xchConn1CB = new QCheckBox(tr("&Exch. conn."), decoderGB);
    xchConn1CB->setObjectName("xch1");
    decdataLayout->addWidget(xchConn1CB, 2, 2);

    /*spacer to push contents of box to top */
    decoderGBL->addStretch();


    QDialogButtonBox* buttonBox = new QDialogButtonBox(
            QDialogButtonBox::Ok |
            QDialogButtonBox::Cancel);

    connect(buttonBox, SIGNAL(accepted()), this, SLOT(validate()));
    connect(buttonBox, SIGNAL(rejected()), this,
            SLOT(reject()));
    baseLayout->addWidget(buttonBox);
}


GaMessage::Protocol DriveDialog::getProtocol()
{
    return (GaMessage::Protocol)protocolBG->checkedId();
}


void DriveDialog::setProtocol(GaMessage::Protocol protocol)
{
    QAbstractButton *button;

    if (protocol == GaMessage::proNone) {
        button = protocolBG->button(pref.protocol);
        if (button != NULL)
            button->setChecked(true);
    }
    else {
        button = protocolBG->button(protocol);
        if (button != NULL)
            button->setChecked(true);
    }
    updateValidators();
}


void DriveDialog::setActiveTime(int atime)
{
    if (atime != -1)
        activeTimeSB->setValue(atime);
    else
        activeTimeSB->setValue(pref.activetime);
}


int DriveDialog::getActiveTime()
{
    return activeTimeSB->value();
}


int DriveDialog::getSRCPBus1()
{
    return srcpBus1LE->text().toInt();
}


void DriveDialog::setSRCPBus1(int bus)
{
    srcpBus1LE->setText(QString::number(bus));
}


int DriveDialog::getAddress1()
{
    return address1LE->text().toInt();
}


void DriveDialog::setAddress1(int addr)
{
    address1LE->setText(QString::number(addr));
}


void DriveDialog::setPort1(int aport)
{
    port1SB->setValue(aport);
}


int DriveDialog::getPort1()
{
    return port1SB->value();
}


int DriveDialog::getXChangeConn1()
{
    return xchConn1CB->isEnabled() ?
        (xchConn1CB->isChecked() ? 1 : 0) : -1;
}


void DriveDialog::setXChangeConn1(int xch)
{
    if (xch == -1)
        xchConn1CB->setChecked(false);
    else
        xchConn1CB->setChecked(xch);
}

void DriveDialog::slotProtocolChanged(int)
{
    updateValidators();
}


void DriveDialog::updateValidators()
{
    /*
     *  id                          protocol
     *  -------------------------------------
     *  -1  none
     *   GaMessage::proMM           MM
     *   GaMessage::proDCC          DCC
     *   GaMessage::proSelectrix    Selectrix
     *   GaMessage::proServer       Server
     *  -------------------------------------
     */

    int prot = protocolBG->checkedId();

    switch (prot) {
        case GaMessage::proMM:
            // MM
            addressVdt->setTop(MAX_GAMM);
            updateAddressTooltip();

            port1Label->setEnabled(false);
            port1SB->setEnabled(false);
            // new range for port spinboxes
            port1SB->setMinimum(MINMMPORT);
            port1SB->setMaximum(MAXMMPORT);

            xchConn1CB->setEnabled(true);

            if (address1LE->text().toInt() > MAX_GAMM)
                address1LE->setText(QString::number(MAX_GAMM));

            break;

        case GaMessage::proDCC:
            //DCC
            addressVdt->setTop(MAX_GADCC);
            updateAddressTooltip();

            port1Label->setEnabled(false);
            port1SB->setEnabled(false);
            // new range for port spinboxes
            port1SB->setMinimum(MINDCCPORT);
            port1SB->setMaximum(MAXDCCPORT);

            xchConn1CB->setEnabled(true);

            break;

        case GaMessage::proSelectrix:
            //Selectrix
            addressVdt->setTop(MAX_GASX);
            updateAddressTooltip();

            // new range for port spinboxes
            port1SB->setMinimum(MINSXPORT);
            port1SB->setMaximum(MAXSXPORT);

            //xchConn1CB->setEnabled(false);

            port1Label->setEnabled(true);
            port1SB->setEnabled(true);

            break;

        case GaMessage::proServer:
            // Server
            // MAGIC: Validator limits for Protocol by Server
            addressVdt->setTop(9999);
            updateAddressTooltip();

            // new range for port spinboxes
            port1SB->setMinimum(MINSVPORT);
            port1SB->setMaximum(MAXSVPORT);

            //xchConn1CB->setEnabled(false);

            port1Label->setEnabled(true);
            port1SB->setEnabled(true);

            break;

        default:
            // undefined protocol
            break;
    }
}

void DriveDialog::updateAddressTooltip()
{
    address1LE->setToolTip(tr(
                "Enter decoder address.\n"
                "Valid range is %1..%2.")
            .arg(addressVdt->bottom())
            .arg(addressVdt->top()));
}

void DriveDialog::validate()
{
    if (address1LE->isEnabled() && !address1LE->hasAcceptableInput()) {
        address1LE->setFocus();
        address1LE->selectAll();
        QMessageBox::warning(this, tr("Unvalid address detected"),
                tr("Value of decoder address is not valid.\n"
                    "Valid range is %1..%2.")
                .arg(addressVdt->bottom())
                .arg(addressVdt->top())
        , QMessageBox::Ok);
        return;
    }
    accept();
}

