/***************************************************************************
                           routelistwindow.cpp
                           -------------------
    copyright            : (C) 2007, 2009 by Guido Scholz
    email                : guido.scholz@bayernline.de
    last modified        : $Date: 2009-03-07 12:49:36 $
                           $Revision: 1.10 $
****************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
   This code creates a dockable window with a routing list
 ***************************************************************************/

#include "route.h"
#include "routelistwindow.h"
#include "routelvi.h"


RouteListWindow::RouteListWindow(QWidget* parent, const char* name,
        Router* router): QDockWidget(tr("Routes"), parent),
    gbsRouter(router), visualMode(kvmNormal)
{
    setObjectName(name);
    setFeatures(DockWidgetClosable | DockWidgetMovable | DockWidgetFloatable);
    //setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
    //setResizeEnabled(true);

    routeLV = new RouteListView(this, "routeListView");
    Q_CHECK_PTR(routeLV);
    setWidget(routeLV);

    connect(routeLV, SIGNAL(currentItemChanged(QTreeWidgetItem*,
                    QTreeWidgetItem*)),
            this, SLOT(currentRouteChanged(QTreeWidgetItem*,
                    QTreeWidgetItem*)));

    connect(routeLV, SIGNAL(itemActivated(QTreeWidgetItem*, int)),
            this, SLOT(slotEditRouteAt(QTreeWidgetItem*, int)));

    connect(routeLV, SIGNAL(spacePressed(QTreeWidgetItem*)),
           this, SLOT(slotToggleRouteState(QTreeWidgetItem*)));

    connect(routeLV, SIGNAL(deletePressed(QTreeWidgetItem*)),
           this, SLOT(slotDeleteRoute(QTreeWidgetItem*)));

    connect(routeLV, SIGNAL(enterPressed(QTreeWidgetItem*)),
           this, SLOT(slotEditRoute(QTreeWidgetItem*)));

    connect(routeLV, SIGNAL(insertPressed()),
            this, SLOT(slotRouteAdd()));
}


/**
 * the router triggers this slot when
 *   - a new route element is recorded/modified
 * */
void RouteListWindow::updateRouteData(Route* rt)
{
    if (routeLV == NULL)
        return;

    RouteLVI* lvi = routeLV->getRouteLVIByRoute(rt);

    if (lvi != NULL)
        lvi->updateRouteData();
}


/**
 * the router triggers this slot when
 *   - the state of a route changes
 * */
void RouteListWindow::updateRouteState(Route* rt)
{
    if (rt == NULL)
        return;

    RouteLVI* lvi = routeLV->getRouteLVIByRoute(rt);

    if (lvi != NULL) {
        lvi->updateRouteStatePixmap();

        // if row is selected also update menu buttons
        if (lvi == routeLV->currentItem())
            emit selectedRouteChangedState();
    }
}

/**
 * the router triggers this slot when
 *   - a new file is loaded
 *   - all routes are cleared
 * */
void RouteListWindow::updateRouteList()
{
    if (gbsRouter == NULL)
        return;

    Route* rt = NULL;
    routeLV->clear();
    emit routeListIsEmpty();

    unsigned int routecount = gbsRouter->getRouteCount();

    for (unsigned int row = 0; row < routecount; row++) {
        rt = gbsRouter->getRouteAt(row);
        if (rt != NULL)
            //RouteLVI* nlvi = new RouteLVI(routeLV, rt);
            new RouteLVI(routeLV, rt);
    }
    if (routecount > 0) {
        routeLV->setCurrentItem(routeLV->topLevelItem(0));
        currentRouteChanged(routeLV->topLevelItem(0), NULL);
        //routeLV->setCurrentItem(routeLV->firstChild());
        //currentRouteChanged(routeLV->firstChild());
    }
}


/**
 * Triggered by main window when the user activates a route.
 * */
void RouteListWindow::slotRouteActivate()
{
    if (gbsRouter == NULL || visualMode == kvmEditLayout)
        return;

    RouteLVI* lvi = (RouteLVI*) routeLV->currentItem();

    if (lvi != NULL)
        gbsRouter->activateRoute(lvi->getRoute());
}


/**
 * Triggerd by route list view when route toggling is done by keyboard.
 * */
void RouteListWindow::slotToggleRouteState(QTreeWidgetItem * lvi)
{
    if (gbsRouter == NULL || visualMode == kvmEditLayout)
        return;

    Route* sr = static_cast<RouteLVI*>(lvi)->getRoute();

    if (sr != NULL) {
        if (sr->getState() != Route::rsUnlocked)
            slotWithdrawRoute(sr);
        else
            slotActivateRoute(sr);
    }
}


/**
 * internal use for route state toggling
 */
void RouteListWindow::slotActivateRoute(Route* sr)
{
    if (gbsRouter == NULL || visualMode == kvmEditLayout)
        return;

    gbsRouter->activateRoute(sr);
}


/**
 * Triggered by main window when the user releases a route.
 * */
void RouteListWindow::slotRouteRelease()
{
    if (gbsRouter == NULL || visualMode == kvmEditLayout)
        return;

    RouteLVI* lvi = (RouteLVI*) routeLV->currentItem();

    if (lvi != NULL)
        slotReleaseRoute(lvi->getRoute());
}


/**
 * internal use
 */
void RouteListWindow::slotReleaseRoute(Route* sr)
{
    if (gbsRouter == NULL || visualMode == kvmEditLayout)
        return;

    gbsRouter->releaseRoute(sr);
}


/**
 * Triggered by main window when the user withdraws a route.
 * */
void RouteListWindow::slotRouteWithdraw()
{
    if (gbsRouter == NULL || visualMode == kvmEditLayout)
        return;

    RouteLVI* lvi = (RouteLVI*) routeLV->currentItem();

    if (lvi != NULL)
        slotWithdrawRoute(lvi->getRoute());
}


/**
 * internal use
 */
void RouteListWindow::slotWithdrawRoute(Route* sr)
{
    if (gbsRouter == NULL || visualMode == kvmEditLayout)
        return;

    gbsRouter->withdrawRoute(sr);
}


/**
 * Triggered by main window when the user adds a new route.
 * */
void RouteListWindow::slotRouteAdd()
{
    if (visualMode != kvmEditRoute)
        return;

    Route* nr = gbsRouter->addNewRoute();

    if (nr != NULL) {
        RouteLVI* nlvi = new RouteLVI(routeLV, nr);
        routeLV->setCurrentItem(nlvi);
        currentRouteChanged(nlvi, NULL);

        // tell main window that list now is _not_ empty
        if (routeLV->topLevelItemCount() == 1)
            emit routeListIsEmpty();
    }
}


/**
 * Triggered by main window when the user edits the selected route.
 * */
void RouteListWindow::slotRouteEdit()
{
    if (visualMode == kvmEditRoute)
        slotEditRoute(routeLV->currentItem());
}


/**
 * Used by
 *   - route table when route edit ist triggered by keyboard.
 *   - main window via slotRouteEdit()
 * */
void RouteListWindow::slotEditRoute(QTreeWidgetItem* lvi)
{
    if (visualMode == kvmEditRoute) {
        if (lvi == NULL)
            return;

        if (gbsRouter->editRoute(parentWidget(),
                    static_cast<RouteLVI*>(lvi)->getRoute()))
            static_cast<RouteLVI*>(lvi)->updateRouteData();
    }
}


void RouteListWindow::slotEditRouteAt(QTreeWidgetItem* lvi, int)
{
    slotEditRoute(lvi);
}

/**
 * Triggered by main window when the user copies the selected route.
 * */
void RouteListWindow::slotRouteCopy()
{
    if (visualMode == kvmEditRoute) {
        RouteLVI* lvi = (RouteLVI*) routeLV->currentItem();
        if (lvi != NULL) {
            Route* cr = lvi->getRoute();
            if (cr != NULL) {
                Route* nr = gbsRouter->copyRoute(cr);
                if (nr != NULL) {
                    RouteLVI* nlvi = new RouteLVI(routeLV, nr);
                    nlvi->setSelected(true);
                }
            }
        }
    }
}


/**
 * Triggered by main window when the user deletes the current route.
 * */
void RouteListWindow::slotRouteDelete()
{
    slotDeleteRoute(routeLV->currentItem());
}


void RouteListWindow::slotDeleteRoute(QTreeWidgetItem * lvi)
{
    if (visualMode == kvmEditRoute) {
        /*TODO: ask for "Do you realy want to delete this route?*/

        if (lvi != NULL) {
            gbsRouter->deleteRoute(static_cast<RouteLVI*>(lvi)->getRoute());
            routeLV->takeTopLevelItem(routeLV->indexOfTopLevelItem(lvi));
            delete lvi;

            // if number of routes is null send signal to main window
            // to update edit button states
            if (routeLV->topLevelItemCount() == 0)
                emit routeListIsEmpty();
        }
    }
}

/**
 * QListview emits this signal for every list selection change:
 * void QListView::selectionChanged(QTreeWidgetItem *)
 *
 *   - send current route state to main window to
 *     update toolbar and menu items
 */
void RouteListWindow::currentRouteChanged(QTreeWidgetItem* lvi,
        QTreeWidgetItem* previous)
{
    if (lvi != NULL) {
        Route* cr = static_cast<RouteLVI*>(lvi)->getRoute();

        if (cr != NULL) {
            // send route state to main window
            emit selectedRouteChangedState();

            // tell router to clear highlighting of last selected route
            gbsRouter->selectedRouteChanged(cr);
        }
    }
}

/**
 * Triggered by main window when the user switches between
 * different visual modes.
 * */
void RouteListWindow::switchVisualMode(elemVisualMode vm)
{
    if (visualMode != vm) {
        visualMode = vm;
        if (visualMode == kvmEditRoute)
            currentRouteChanged(routeLV->currentItem(), NULL);
    }
}


bool RouteListWindow::hasCurrentItem()
{
    return (routeLV->currentItem() != NULL);
}


int RouteListWindow::getCurrentItemState()
{
    RouteLVI* lvi = (RouteLVI*) routeLV->currentItem();
    if (lvi != NULL) {
        Route* cr = lvi->getRoute();
        if (cr != NULL)
            return cr->getState();
    }
    return -1;
}

